@@ -1,6 +1,6 @@
 // MPTable generation (on emulators)
 //
-// Copyright (C) 2008,2009  Kevin O'Connor <kevin@koconnor.net>
+// Copyright (C) 2008-2010  Kevin O'Connor <kevin@koconnor.net>
 // Copyright (C) 2006 Fabrice Bellard
 //
 // This file may be distributed under the terms of the GNU LGPLv3 license.
@@ -21,7 +21,7 @@ mptable_init(void)
     dprintf(3, "init MPTable\n");
 
     // Config structure in temp area.
-    struct mptable_config_s *config = malloc_tmphigh(32*1024);
+    struct mptable_config_s *config = malloc_tmp(32*1024);
     if (!config) {
         warn_noalloc();
         return;
@@ -59,7 +59,7 @@ mptable_init(void)
         cpu->apicid = i;
         cpu->apicver = apic_version;
         /* cpu flags: enabled, bootstrap cpu */
-        cpu->cpuflag = (i < CountCPUs) | ((i == 0) << 1);
+        cpu->cpuflag = ((i<CountCPUs) ? 0x01 : 0x00) | ((i==0) ? 0x02 : 0x00);
         cpu->cpusignature = cpuid_signature;
         cpu->featureflag = cpuid_features;
         cpu++;
@@ -67,10 +67,11 @@ mptable_init(void)
     int entrycount = cpu - cpus;
 
     // PCI buses
-    struct mpt_bus *bus = (void*)cpu;
-    int bdf, max, lastbus = -1;
-    foreachpci(bdf, max) {
-        int curbus = pci_bdf_to_bus(bdf);
+    struct mpt_bus *buses = (void*)cpu, *bus = buses;
+    int lastbus = -1;
+    struct pci_device *pci;
+    foreachpci(pci) {
+        int curbus = pci_bdf_to_bus(pci->bdf);
         if (curbus == lastbus)
             continue;
         lastbus = curbus;
@@ -79,7 +80,6 @@ mptable_init(void)
         bus->busid = curbus;
         memcpy(bus->bustype, "PCI   ", sizeof(bus->bustype));
         bus++;
-        entrycount++;
     }
 
     /* isa bus */
@@ -88,11 +88,12 @@ mptable_init(void)
     bus->type = MPT_TYPE_BUS;
     isabusid = bus->busid = lastbus + 1;
     memcpy(bus->bustype, "ISA   ", sizeof(bus->bustype));
-    entrycount++;
+    bus++;
+    entrycount += bus - buses;
 
     /* ioapic */
     u8 ioapic_id = CountCPUs;
-    struct mpt_ioapic *ioapic = (void*)&bus[1];
+    struct mpt_ioapic *ioapic = (void*)bus;
     memset(ioapic, 0, sizeof(*ioapic));
     ioapic->type = MPT_TYPE_IOAPIC;
     ioapic->apicid = ioapic_id;
@@ -106,7 +107,8 @@ mptable_init(void)
     int dev = -1;
     unsigned short mask = 0, pinmask = 0;
 
-    foreachpci(bdf, max) {
+    foreachpci(pci) {
+        u16 bdf = pci->bdf;
         int pin = pci_config_readb(bdf, PCI_INTERRUPT_PIN);
         int irq = pci_config_readb(bdf, PCI_INTERRUPT_LINE);
         if (pin == 0)
@@ -151,7 +153,6 @@ mptable_init(void)
         }
         intsrc++;
     }
-    entrycount += intsrc - intsrcs;
 
     /* Local interrupt assignment */
     intsrc->type = MPT_TYPE_LOCAL_INT;
@@ -162,7 +163,6 @@ mptable_init(void)
     intsrc->dstapic = 0; /* BSP == APIC #0 */
     intsrc->dstirq = 0; /* LINTIN0 */
     intsrc++;
-    entrycount++;
 
     intsrc->type = MPT_TYPE_LOCAL_INT;
     intsrc->irqtype = 1; /* NMI */
@@ -172,7 +172,7 @@ mptable_init(void)
     intsrc->dstapic = 0; /* BSP == APIC #0 */
     intsrc->dstirq = 1; /* LINTIN1 */
     intsrc++;
-    entrycount++;
+    entrycount += intsrc - intsrcs;
 
     // Finalize config structure.
     int length = (void*)intsrc - (void*)config;
